@@ -1,7 +1,8 @@
 #! /usr/bin/env python3
 
 '''
-Build+test script for PyMuPDF, mostly for use with github builds.
+Build+test script for PyMuPDF using cibuildwheel. Mostly for use with github
+builds but can also be used to build Pyodide wheels.
 
 We run cibuild manually, in order to build and test the two wheel
 flavours that make up PyMuPDF:
@@ -11,7 +12,7 @@ flavours that make up PyMuPDF:
         libraries for the MuPDF C and C++ bindings.
     PyMuPDF
         Specific to particular versions of Python. Contains the rest of
-        the(classic and rebased) PyMuPDF implementation.
+        the PyMuPDF implementation.
 
 Args:
     build
@@ -19,32 +20,42 @@ Args:
     build-devel
         Build using cibuild with `--platform` set.
     pip_install <prefix>
-        Run `pip install <prefix>-*<platform_tag>.whl`,
+        For internal use. Runs `pip install <prefix>-*<platform_tag>.whl`,
         where `platform_tag` will be things like 'win32', 'win_amd64',
         'x86_64`, depending on the python we are running on.
     venv
-        Run remaining args inside venv.
+        Run with remaining args inside a venv.
     test
         Internal.
 
-We look at these items in the environment, should be unset (treated as '0'),
-'0' or '1'.
+We also look at specific items in the environment, should be unset (treated as
+'0'), '0' or '1'. We use environment variables here to allow use with Github
+action inputs, which can't be easily translated into command-line arguments.
 
     inputs_flavours
         If unset or '1', build separate PyMuPDF and PyMuPDFb wheels.
         If '0', build complete PyMuPDF wheels.
+    inputs_sdist
     inputs_skeleton
         Build minimal wheel; for testing only.
-    inputs_sdist
+    inputs_wheels_cps:
+        Python versions to build for. E.g. 'cp38* cp312*'.
     inputs_wheels_default
+        Default value for other inputs_wheels_*.
     inputs_wheels_linux_aarch64
     inputs_wheels_linux_auto
     inputs_wheels_linux_pyodide
     inputs_wheels_macos_arm64
     inputs_wheels_macos_auto
     inputs_wheels_windows_auto
+    inputs_PYMUPDF_SETUP_MUPDF_BUILD
+        Used to directly set PYMUPDF_SETUP_MUPDF_BUILD.
+        E.g. 'git:--recursive --depth 1 --shallow-submodules --branch master https://github.com/ArtifexSoftware/mupdf.git'
+    inputs_wheels_implementations
+        Used to directly set PYMUPDF_SETUP_IMPLEMENTATIONS.
+        'a', 'b', 'ab'.
 
-Buiding for Pyodide
+Building for Pyodide
 
     If `inputs_wheels_linux_pyodide` is true and we are on Linux, we clone
     `emsdk.git`, set it up, and run `pyodide build`. This runs our setup.py
@@ -53,20 +64,21 @@ Buiding for Pyodide
 
 Example usage:
 
-     PYMUPDF_SETUP_MUPDF_BUILD=../mupdf-master py -3.9-32 PyMuPDF/scripts/gh_release.py venv build-devel
-   
+     PYMUPDF_SETUP_MUPDF_BUILD=../mupdf py -3.9-32 PyMuPDF/scripts/gh_release.py venv build-devel
 '''
 
-
 import glob
+import inspect
 import os
 import platform
 import shlex
-import sys
 import subprocess
+import sys
 import textwrap
 
 
+pymupdf_dir = os.path.abspath( f'{__file__}/../..')
+
 def main():
 
     log( '### main():')
@@ -89,6 +101,7 @@ def main():
         v = os.environ[ k]
         log( f'    {k}: {v!r}')
     
+    valgrind = False
     if len( sys.argv) == 1:
         args = iter( ['build'])
     else:
@@ -99,7 +112,7 @@ def main():
         except StopIteration:
             break
         if arg == 'build':
-            build()
+            build(valgrind=valgrind)
         elif arg == 'build-devel':
             if platform.system() == 'Linux':
                 p = 'linux'
@@ -130,12 +143,14 @@ def main():
         elif arg == 'test':
             project = next(args)
             package = next(args)
-            test( project, package)
+            test( project, package, valgrind=valgrind)
+        elif arg == '--valgrind':
+            valgrind = int(next(args))
         else:
             assert 0, f'Unrecognised {arg=}'
 
 
-def build( platform_=None): 
+def build( platform_=None, valgrind=False): 
     log( '### build():')   
     
     platform_arg = f' --platform {platform_}' if platform_ else ''
@@ -165,6 +180,7 @@ def build( platform_=None):
     inputs_wheels_windows_auto = get_bool('inputs_wheels_windows_auto', inputs_wheels_default)
     inputs_wheels_cps = os.environ.get('inputs_wheels_cps')
     inputs_PYMUPDF_SETUP_MUPDF_BUILD = os.environ.get('inputs_PYMUPDF_SETUP_MUPDF_BUILD')
+    inputs_wheels_implementations = os.environ.get('inputs_wheels_implementations', 'b')
     
     log( f'{inputs_flavours=}')
     log( f'{inputs_sdist=}')
@@ -182,7 +198,7 @@ def build( platform_=None):
     # Build Pyodide wheel if specified.
     #
     if platform.system() == 'Linux' and inputs_wheels_linux_pyodide:
-        build_pyodide_wheel()
+        build_pyodide_wheel(inputs_wheels_implementations, inputs_PYMUPDF_SETUP_MUPDF_BUILD)
     
     # Build 
     #
@@ -205,25 +221,14 @@ def build( platform_=None):
                 ret.append(item)
         return ' '.join(ret)
     
-    cps = inputs_wheels_cps if inputs_wheels_cps else 'cp38* cp39* cp310* cp311*'
+    cps = inputs_wheels_cps if inputs_wheels_cps else 'cp38* cp39* cp310* cp311* cp312*'
     set_if_unset( 'CIBW_BUILD', cps)
     
-    if 0:
-        cps = ([]
-                + inputs_wheels_cp38 * ['cp38*']
-                + inputs_wheels_cp39 * ['cp39*']
-                + inputs_wheels_cp310 * ['cp310*']
-                + inputs_wheels_cp311 * ['cp311*']
-                )
-        cps = ' '.join(cps)
-        log(f'{cps=}')
-        set_if_unset( 'CIBW_BUILD', cps)
-    
     if platform.system() == 'Linux':
         set_if_unset(
                 'CIBW_ARCHS_LINUX',
                 make_string(
-                    'auto' * inputs_wheels_linux_auto,
+                    'auto64' * inputs_wheels_linux_auto,
                     'aarch64' * inputs_wheels_linux_aarch64,
                     ),
                 )
@@ -254,19 +259,30 @@ def build( platform_=None):
             log(f'Not running cibuildwheel because CIBW_ARCHS_MACOS is empty string.')
             return
     
-    def env_set(name, value, pass_=False):
-        assert isinstance( value, str)
-        if not name.startswith('CIBW'):
-            assert pass_, f'{name=} {value=}'
-        env_extra[ name] = value
-        if pass_ and platform.system() == 'Linux':
+    def env_pass(name):
+        '''
+        Adds `name` to CIBW_ENVIRONMENT_PASS_LINUX if required to be available
+        when building wheel with cibuildwheel.
+        '''
+        if platform.system() == 'Linux':
             v = env_extra.get('CIBW_ENVIRONMENT_PASS_LINUX', '')
             if v:
                 v += ' '
             v += name
             env_extra['CIBW_ENVIRONMENT_PASS_LINUX'] = v
+    
+    def env_set(name, value, pass_=False):
+        assert isinstance( value, str)
+        if not name.startswith('CIBW'):
+            assert pass_, f'{name=} {value=}'
+        env_extra[ name] = value
+        if pass_:
+            env_pass(name)
 
-    env_set('PYMUPDF_SETUP_IMPLEMENTATIONS', 'ab', pass_=1)
+    if os.environ.get('PYMUPDF_SETUP_LIBCLANG'):
+        env_pass('PYMUPDF_SETUP_LIBCLANG')
+    
+    env_set('PYMUPDF_SETUP_IMPLEMENTATIONS', inputs_wheels_implementations, pass_=1)
     if inputs_skeleton:
         env_set('PYMUPDF_SETUP_SKELETON', inputs_skeleton, pass_=1)
     
@@ -277,13 +293,11 @@ def build( platform_=None):
     
     def set_cibuild_test():
         log( f'set_cibuild_test(): {inputs_skeleton=}')
-        if inputs_skeleton:
-            env_set('CIBW_TEST_COMMAND', 'python {project}/scripts/gh_release.py test {project} {package}')
-        else:
-            env_set('CIBW_TEST_REQUIRES', 'fontTools pytest')
-            env_set('CIBW_TEST_COMMAND', 'python {project}/tests/run_compound.py pytest -s {project}/tests')
+        valgrind_text = ''
+        if valgrind:
+            valgrind_text = ' --valgrind 1'
+        env_set('CIBW_TEST_COMMAND', f'python {{project}}/scripts/gh_release.py{valgrind_text} test {{project}} {{package}}')
     
-    pymupdf_dir = os.path.abspath( f'{__file__}/../..')
     if pymupdf_dir != os.path.abspath( os.getcwd()):
         log( f'Changing dir to {pymupdf_dir=}')
         os.chdir( pymupdf_dir)
@@ -302,8 +316,8 @@ def build( platform_=None):
         run( 'echo after flavour=b')
         run( 'ls -l wheelhouse')
 
-        # Now build PyMuPDF wheels. cibuildwheel will build one for each
-        # Python version.
+        # Now set environment to build PyMuPDF wheels. cibuildwheel will build
+        # one for each Python version.
         #
         
         # Tell cibuildwheel not to use `auditwheel`, because it cannot cope
@@ -342,7 +356,7 @@ def build( platform_=None):
     run( 'ls -lt wheelhouse')
 
 
-def build_pyodide_wheel():
+def build_pyodide_wheel( implementations, inputs_PYMUPDF_SETUP_MUPDF_BUILD):
     '''
     Build Pyodide wheel.
 
@@ -360,11 +374,14 @@ def build_pyodide_wheel():
     # Disable libcrypto because not available in Pyodide.
     env_extra['HAVE_LIBCRYPTO'] = 'no'
 
+    # Set PYMUPDF_SETUP_MUPDF_BUILD.
+    if inputs_PYMUPDF_SETUP_MUPDF_BUILD:
+        env_extra['PYMUPDF_SETUP_MUPDF_BUILD'] = inputs_PYMUPDF_SETUP_MUPDF_BUILD
+
     # Tell MuPDF to build for Pyodide.
     env_extra['OS'] = 'pyodide'
 
-    # Build only classic PyMuPDF.
-    env_extra['PYMUPDF_SETUP_IMPLEMENTATIONS'] = 'a'
+    env_extra['PYMUPDF_SETUP_IMPLEMENTATIONS'] = implementations
     
     # Build a single wheel without a separate PyMuPDFb wheel.
     env_extra['PYMUPDF_SETUP_FLAVOUR'] = 'pb'
@@ -391,7 +408,15 @@ def build_pyodide_wheel():
     run('ls -l wheelhouse/')
 
 
-def venv( command=None, packages=None):
+def cpu_bits():
+    return 32 if sys.maxsize == 2**31 - 1 else 64
+
+
+# Name of venv used by `venv()`.
+#
+venv_name = f'venv-pymupdf-{platform.python_version()}-{cpu_bits()}'
+
+def venv( command=None, packages=None, quick=False):
     '''
     Runs remaining args, or the specified command if present, in a venv.
     
@@ -400,19 +425,39 @@ def venv( command=None, packages=None):
         to run the venv's python.
     packages:
         List of packages (or comma-separated string) to install.
+    quick:
+        If true and venv directory already exists, we don't recreate venv or
+        install Python packages in it.
     '''
-    venv_name = f'venv-pymupdf-{platform.python_version()}'
     command2 = ''
-    command2 += f'{sys.executable} -m venv {venv_name}'
+    ssp = ''
+    if platform.system() == 'OpenBSD':
+        # libclang not available from pypi.org, but system py3-llvm package
+        # works. `pip install` should be run with --no-build-isolation and
+        # explicit `pip install swig setuptools psutil`.
+        ssp = ' --system-site-packages'
+        log(f'OpenBSD: libclang not available from pypi.org.')
+        log(f'OpenBSD: system package `py3-llvm` must be installed.')
+        log(f'OpenBSD: creating venv with --system-site-packages.')
+        log(f'OpenBSD: `pip install .../PyMuPDF` must be preceded by install of swig etc.')
+    if quick and os.path.isdir(venv_name):
+        log(f'{quick=}: Not creating venv because directory already exists: {venv_name}')
+        command2 += 'true'
+    else:
+        quick = False
+        command2 += f'{sys.executable} -m venv{ssp} {venv_name}'
     if platform.system() == 'Windows':
         command2 += f' && {venv_name}\\Scripts\\activate'
     else:
         command2 += f' && . {venv_name}/bin/activate'
-    if packages:
+    if quick:
+        log(f'{quick=}: Not upgrading pip or installing packages.')
+    else:
         command2 += ' && python -m pip install --upgrade pip'
-        if isinstance(packages, str):
-            packages = packages.split(',')
-        command2 += ' && pip install ' + ' '.join(packages)
+        if packages:
+            if isinstance(packages, str):
+                packages = packages.split(',')
+            command2 += ' && pip install ' + ' '.join(packages)
     command2 += ' &&'
     if isinstance( command, str):
         command2 += ' ' + command
@@ -423,41 +468,29 @@ def venv( command=None, packages=None):
     run( command2)
 
 
-def test( project, package):
-
+def test( project, package, valgrind):
     
-    log('### test():')
-    log(f'### test(): {sys.executable=}')
-    log(f'### test(): {project=}')
-    log(f'### test(): {package=}')
-    
-    import fitz
-    import fitz_new
-    print(f'{fitz.bar(3)=}')
-    print(f'{fitz_new.bar(3)=}')
-    
-    return
-    
-    run('ls -l')
-    run( f'ls -l {project}')
-    run( f'ls -l {package}')
-    run( f'ls -l {project}/wheelhouse', check=0)
-    run( f'ls -l {package}/wheelhouse', check=0)
-    
-    wheel_b = glob.glob( f'{project}/wheelhouse/PyMuPDFb-*{platform_tag()}.whl')
-    assert len(wheel_b) == 1, f'{wheel_b=}'
-    wheel_b = wheel_b[0]
-
-    py_version = platform.python_version_tuple()
-    py_version = py_version[:2]
-    py_version = ''.join( py_version)
-    log( '### test(): {py_version=}')
-    wheel_p = glob.glob( f'wheelhouse/PyMuPDF-*-cp{py_version}-*.whl')
-    print(f'{wheel_p=}')
-    #assert len(wheel_p) == 1, f'{wheel_p=}'
-    
-    #run( f'pip install {wheel_b}')
-    #run( f'pip install {wheel_p}')
+    run(f'pip install {test_packages}')
+    if valgrind:
+        log('Installing valgrind.')
+        run(f'sudo apt update')
+        run(f'sudo apt install valgrind')
+        run(f'valgrind --version')
+        
+        log('Running PyMuPDF tests under valgrind.')
+        # We ignore memory leaks.
+        run(
+                f'{sys.executable} {project}/tests/run_compound.py'
+                    f' valgrind --suppressions={project}/valgrind.supp --error-exitcode=100 --errors-for-leak-kinds=none --fullpath-after='
+                    f' pytest {project}/tests'
+                    ,
+                env_extra=dict(
+                    PYTHONMALLOC='malloc',
+                    PYMUPDF_RUNNING_ON_VALGRIND='1',
+                    ),
+                )
+    else:
+        run(f'{sys.executable} {project}/tests/run_compound.py pytest {project}/tests')
     
 
 def pyodide_setup(clean=False):
@@ -578,33 +611,58 @@ def pyodide_setup(clean=False):
     return command
 
 
-def log(text):
-    print(f'{__file__}: {text}')
+if platform.system() == 'Windows':
+    def relpath(path, start=None):
+        try:
+            return os.path.relpath(path, start)
+        except ValueError:
+            # os.path.relpath() fails if trying to change drives.
+            return os.path.abspath(path)
+else:
+    def relpath(path, start=None):
+        return os.path.relpath(path, start)
+
+
+def log(text, caller=0):
+    '''
+    Writes `text` to stdout with prefix showing caller path relative to
+    pymupdf_dir and fn name.
+    '''
+    frame_record = inspect.stack( context=0)[ caller+1]
+    filename    = frame_record.filename
+    line        = frame_record.lineno
+    function    = frame_record.function
+    prefix = f'{relpath(filename, pymupdf_dir)}:{function}(): '
+    print(textwrap.indent(text, prefix))
     sys.stdout.flush()
 
 
-def run(command, env_extra=None, env=None, check=1):
-    if env is None:
-        env = add_env(env_extra)
-    else:
-        assert env_extra is None
-    log(f'Running: {command}')
-    sys.stdout.flush()
-    subprocess.run(command, check=check, shell=1, env=env)
-
-
-def add_env(env_extra):
-    env = os.environ.copy()
+def run(command, env_extra=None, check=1, timeout=None):
+    '''
+    Runs a command using subprocess.run().
+    Args:
+        command:
+            The command to run.
+        env_extra:
+            None or dict containing extra environment variable settings to add
+            to os.environ.
+        check:
+            Whether to raise exception if command fails.
+        timeout:
+            If not None, timeout in seconds; passed directory to
+            subprocess.run(). Note that on MacOS subprocess.run() seems to
+            leave processes running if timeout expires.
+    '''
+    env = None
     if env_extra:
+        env = os.environ.copy()
         env.update(env_extra)
-        log(f'Adding environment:')
-        for n, v in env_extra.items():
-            log(f'    {n}: {v!r}')
-    return env
+    log(f'Running: {command}')
+    return subprocess.run(command, check=check, shell=1, env=env, timeout=timeout)
 
 
 def platform_tag():
-    bits = 32 if sys.maxsize == 2**31 - 1 else 64
+    bits = cpu_bits()
     if platform.system() == 'Windows':
         return 'win32' if bits==32 else 'win_amd64'
     elif platform.system() in ('Linux', 'Darwin'):
@@ -615,5 +673,18 @@ def platform_tag():
         assert 0, f'Unrecognised: {platform.system()=}'
 
 
+test_packages = 'pytest fontTools pymupdf-fonts flake8 pylint'
+if platform.system() == 'Windows' and cpu_bits() == 32:
+    # No pillow wheel available, and doesn't build easily.
+    pass
+else:
+    test_packages += ' pillow'
+if platform.system().startswith('MSYS_NT-'):
+    # psutil not available on msys2.
+    pass
+else:
+    test_packages += ' psutil'
+
+
 if __name__ == '__main__':
     main()
